%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% build dmel bcd rnai atlas  (2013-01-24)
%
%  this makes use of a slightly different version of the alignment code "register_pc.m"
%  which deals with greater variations in eve/ftz stripes present in bcd rnai embyros
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
define_constants; %import useful global constants
subdir = '/home/datasets/bio/bcdrnai_10_7_12/';
genotype = MEL; 
versionstring = '-dmel-rnai'; 
includedensity = false;
vis = false;
forcecoarse = false; % don't force coarse alignment
%forcecoarse = true; % don't force coarse alignment

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% STEP 0: collect metadata and density data (if desired)
% build up summary data, save it and load it in
metadatafile = [subdir 'metadata.mat'];
if fileexists(metadatafile)
  fprintf('loading cached metadata\n');
  load(metadatafile);
else
  fprintf('collecting and saving metadata\n');
  [gene_names,numgenes,genetopc,hybs,numhybs,genetohybind,dyelist] = collect_metadata(subdir,genotype);
  save(metadatafile,'gene_names','numgenes','genotype','genetopc','hybs','numhybs','genetohybind','dyelist');
end;

if (includedensity)
  morphologyfile = [subdir 'morphology_data.mat'];
  if fileexists(morphologyfile)
    fprintf('loading cached morphological data\n');
    load(morphologyfile)
  else
    fprintf('collecting and saving morphological data\n');
    [dmap,vdmap,dcount,stats] = average_densities_robust(subdir,genetopc);
    save(morphologyfile,'dmap','vdmap','dcount','stats');
  end
  for i = 1:numcohorts 
    figure_density(dmap(:,:,i),sprintf('density%s, stage:%s',versionstring,cohort_names{i}),0.025,0.05);
    drawnow;
    print(gcf,sprintf('%s/export/density/density%s-%s.eps',subdir,versionstring,cohort_names{i}),'-depsc2');
  end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% STEP 1: correspondence
D = dir([subdir 'pointclouds.pce/*.pce']);
tic
parfor i = 1:length(D)
  pcfile = [subdir 'pointclouds.pce/' D(i).name(1:end-4) '.pce'];
  corrfile = [subdir 'pointclouds.cor/' D(i).name(1:end-4) '.cor'];
  if ~fileexists(corrfile)
    if (forcecoarse)
      register_pc(pcfile,corrfile,gene_names,versionstring,vis,forcecoarse);
    else
      register_pc_bcdrnai(pcfile,corrfile,gene_names,versionstring,vis,forcecoarse);
    end
  else
    fprintf('skipping %s\n',pcfile);
  end
end
tt = toc;
fprintf('%3.3f seconds for %d embryos = %3.3f seconds/embryo\n',tt,i,(tt/i));


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% STEP 2: calibrating and compositing 
% allow for excluding some set of pointclouds from the build
%
excludelist = {};

% now process the set of point clouds for each (cohort,geneid).
% load in the correspondence files and composite the data into
% the array Emean

% load template geometry
load(sprintf('template/%d_stage5shape.mat',genotype));
numnuclei = size(V{4},1); %TODO: only works for stage 5 cohorts...
D = dir([subdir 'pointclouds.pce/*.pce']);
Emean = zeros(numnuclei,numgenes,numcohorts,numchanneltypes);
Estd = zeros(numnuclei,numgenes,numcohorts,numchanneltypes);
Edev = cell(numgenes,numcohorts,numchanneltypes);
hybmaxave = zeros(numcohorts,numgenes,numhybs,numchanneltypes);
hybmaxavect = zeros(numcohorts,numgenes,numhybs,numchanneltypes);
srclist = {}; srcind = 1;
srclist_bygene = cell(numgenes,numcohorts,numchanneltypes);
uhlist = cell(numgenes,numcohorts,numchanneltypes);
for gg = 1:numgenes %loop over genes 
  for cht = 1:numcohorts %loop over cohort 
    for tt = 1:numchanneltypes %loop over type (e.g. protein,mRNA)
      %running count of pcs averaged into this channel
      ct(gg,cht,tt) = 0;
      pcind = genetopc{gg,cht,tt};
      E = zeros(numnuclei,length(pcind)); %expression for all the genes
      hybindexlist = zeros(1,length(pcind)); %hybridization index for each pc
      srclist_bygene{gg,cht,tt} = {};
      for k = 1:length(pcind)
        %load pointcloud and correspondence file
        fprintf('\rprocessing geneid=%d,cht=%d,type=%s : %d/%d        ',gg,cht,channeltype_names{tt},k,length(pcind))
        pcname = D(pcind(k)).name(1:end-4);
        pcfile = [subdir 'pointclouds.pce/' pcname  '.pce'];
        corrfile = [subdir 'pointclouds.cor/' pcname '.cor'];
        if ~fileexists(corrfile)
          fprintf('\n WARNING: skipping embryo with missing .cor file %s\n',corrfile);
          continue;
        end;
        if ~isempty(strmatch(pcname,excludelist)) 
          fprintf('\n WARNING: skipping embryo from exclude list\n');
          continue;
        end;
        if (dyelist(pcind(k),gg) == dict({'Coumarin'},dye_names))
          fprintf('\n WARNING: skipping Coumarin\n');
          continue;
        end

        srclist_bygene{gg,cht,tt} = union(srclist_bygene{gg,cht,tt},pcname);
        srclist{srcind} = pcname;
        srcind = srcind + 1;
        pc = load_pc(pcfile);
        corr = load_corr(corrfile);
        if (getcohorts(pc) ~= cht), error('pc cohort does not match expected cht value'); end;
        hybindex = find(hybs==pc.hybridization_id);
        if isempty(hybindex), error('pointcloud contains unknown hybridization id, rebuild metadata?'); end;

        %normalize expression for attenuation effects
        sytox = strmatch(pc.nuclear_stain,pc.column);
        [col,info,zerocount] = getgenechannels(pc,gene_names);

        % extract attenuation offset values 
        offset = zeros(size(info,2),1);
        if ~isempty(pc.attenuation_correct)
          for j = 1:size(info,2)
            ind = strmatch(pc.column_info(info(j),1),pc.attenuation_correct);
            if isempty(ind)
              offset(j) = 0;
            else
              offset(j) = pc.attenuation_offset(ind);
            end;
          end;
        end;

        % pull out the channel that actually corresponds to the (gg,tt) we are interested in
        for j = 1:size(col,2)
          type = strmatch(pc.column_info{info(j),2},channeltype_names);
          gene = dict(pc.column_info{info(j),3},gene_names);
          if ((gene==gg) && (type == tt))
            %attenuation correction, set to 1 for the "top" of the embryo nearest the microscope objective
            correction = (pc.data(:,sytox) + offset(j));
            topmost = find((pc.X(:,3)-min(pc.X(:,3)))<2);
            correction = correction / mean(correction(topmost));
            fluor = pc.data(sub2ind(size(pc.data),1:size(pc.data,1),col(:,j)'))';
            ex =  fluor ./ correction;
            %use the matching matrix to take a weighted combination 
            %of measurements.  assumes that the rows of match are
            %L1 normalized
            ct(gg,cht,tt) = ct(gg,cht,tt) + 1;  
            E(:,ct(gg,cht,tt)) = corr.match*ex; 
            hybindexlist(ct(gg,cht,tt)) = hybindex;
          end
        end
      end
 
      % TODO: error checking. make sure E doesn't contain NaNs?
      assert(~any(any(isnan(E))));

      if (ct(gg,cht,tt)>0)
        % now renormalize before averaging
        %scale and offset to minimize stddev
        [Snorm,cgain,coffset] = normalize_signals(E(:,1:ct(gg,cht,tt))');
        Snorm = rerange(Snorm); %range so that mean lies in [0,1]
        if (ct(gg,cht,tt)>1)
          Emean(:,gg,cht,tt) = mean(Snorm)';
          Estd(:,gg,cht,tt) = std(Snorm)';
          Edev{gg,cht,tt} = deverr(Snorm');
        else %with only 1 embryo, we cant estimate stddev
          Emean(:,gg,cht,tt) = Snorm';
          Estd(:,gg,cht,tt) = zeros(size(Snorm'));
          Edev{gg,cht,tt} = 0;
        end;
        fprintf('\n');

        % also compute the per-hybridization averages for this gene
        %  loop over the hybridizations and compute the average for
        %  all pcs in with a particular hybridization index
        uniquehybs = unique(hybindexlist(1:ct(gg,cht,tt)));
        uhybslist{gg,cht,tt} = uniquehybs;
        for h = 1:length(uniquehybs)
          hh = uniquehybs(h);
          select = find(hybindexlist(1:ct(gg,cht,tt))==hh);
          [Snorm,cgain,coffset] = normalize_signals(E(:,select)');
          hybmaxave(cht,gg,hh,tt) = prctile(mean(Snorm),99);
          hybmaxavect(cht,gg,hh,tt) = length(select);
        end
      end
    end
  end
end
save([subdir 'atlas_data.mat'],'Emean','Estd','Edev','ct','genetopc','srclist','hybmaxave','hybmaxavect','uhybslist','srclist_bygene');

% scale expression based on average behavior of hybridizations
% scaled result is relative to an observed timecourse max of 1.
% estimate gains
fprintf('estimating gains, this make take a while...\n');
gainfit = smoothgains(hybmaxave,hybmaxavect,uhybslist,ct,numgenes,gene_names,vis);
maxgainfit = repmat(max(gainfit.*(ct>0),[],2),[1 numcohorts 1]);
sgainfit = gainfit ./ (maxgainfit+(maxgainfit==0));

% now do the actual scaling
Emean_scaled = zeros(size(Emean));
for tt = 1:numchanneltypes
  for gg = 1:numgenes
    for cht = 1:numcohorts 
      unscaled = Emean(:,gg,cht,tt);
      p = prctile(unscaled,99); %99 prctile
      Emean_scaled(:,gg,cht,tt) = unscaled*sgainfit(gg,cht,tt)/(p+(p==0));
    end 
  end
end

% cache the results as a .mat file
save([subdir 'atlas_data.mat'],'Emean','Emean_scaled','Estd','Edev','ct','genetopc','gainfit','srclist','hybmaxave','hybmaxavect','uhybslist','srclist_bygene');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% STEP 4: write out atlas file 
%
load(sprintf('template/%d_stage5shape.mat',genotype));
for cht = 1:3
  V{cht} = V{4};
end
for cht = 10:11
  V{cht} = V{9};
end

exportdir = [subdir 'export'];
mrnalist = gene_names;
proteinlist =  gene_names;
% NOTE: if you only write out a subset of genes
% you will also need to modify srclistfinal.
srclistfinal = unique(srclist);
% pull out the subset from gene_names along
% with their indicies.
subset = [];
gene_names2 = [];
types = [];
mrnasubset = [];
for i = 1:length(mrnalist)
  mrnasubset(i) = strmatch(mrnalist{i},gene_names,'exact');
  gene_names2{i} = gene_names{mrnasubset(i)};
  types{i} = 'mRNA';
end;
offset = length(mrnalist);
proteinsubset = [];
for i = 1:length(proteinlist)
  proteinsubset(i) = strmatch(proteinlist{i},gene_names,'exact');
  gene_names2{i+offset} = gene_names{proteinsubset(i)};
  types{i+offset} = 'protein';
end;

% write out atlas file
clear X chort embryospergene;
X = cell(1,numcohorts);
embryospergene = zeros(numcohorts,length(gene_names2));
for cht = 1:numcohorts 
  X{cht} = V{cht}; %copy in template geometry
  if (includedensity) %only add density channel if desired.
    Ex{cht} = [Emean_scaled(:,mrnasubset,cht,1) Emean_scaled(:,proteinsubset,cht,2) density(:,cht)];
    embryospergene(cht,:) = [ct(mrnasubset,cht,1)' ct(proteinsubset,cht,2)' dcount(cht)];
  else
    Ex{cht} = [Emean_scaled(:,mrnasubset,cht,1) Emean_scaled(:,proteinsubset,cht,2)];
    embryospergene(cht,:) = [ct(mrnasubset,cht,1)' ct(proteinsubset,cht,2)'];
  end;
end;
nuclear_count = size(X{1},1);
goodstages = sum(embryospergene,2)>0;
save_atlas(sprintf('%s/atlas.vpc',exportdir),X,Ex,nnbrs{4},nbrs{4},gene_names2,types,cohort_names,cohort_times,embryospergene,srclistfinal);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% STEP 5: write out visual summary to upload to BID
%
% generate gene expression pattern thumbnails 
Exu = cell(numgenes,numcohorts,numchanneltypes);
res = 150 
for tt = 1:numchanneltypes
  for cht = 1:numcohorts 
    VV = V{cht};
    VV(:,1) = VV(:,1)-min(VV(:,1));
    VV(:,1) = VV(:,1)/max(VV(:,1));
    for gg = 1:numgenes
      if (ct(gg,cht,tt)>0)
        S = render_signal(VV,Emean_scaled(:,gg,cht,tt),res,'cubic');
        S = conv2(S,fspecial('gaussian',10,1),'same');
        Exu{gg,cht,tt}(:,:) = S;
        Exu{gg,cht,tt}(1,:) = 0; Exu{gg,cht,tt}(end,:) = 0;
        Exu{gg,cht,tt}(:,1) = 0; Exu{gg,cht,tt}(:,end) = 0;
        if (vis)
          figure(1); imagesc(Exu{gg,cht,tt}); axis image;
          caxis([0 1]); drawnow;
        end;
        fprintf('%d,%d,%d\n',gg,tt,cht);
      else
        Exu{gg,cht,tt} = zeros(res,res);
      end;
    end
  end
end
save([subdir 'atlas_data_unrolled.mat'],'Exu');

%
% write out half thumbnails for use in database web interface
%
for tt = 1:numchanneltypes
  if (tt==MRNA)
    subset = mrnasubset;
  else
    subset = proteinsubset;
  end
  for gg = 1:length(subset)
    g = subset(gg);
    for cht = 1:numcohorts 
      if (ct(g,cht,tt)>0)
        thumb = 0.5*(flipud(Exu{g,cht,tt}(76:150,:))+Exu{g,cht,tt}(1:75,:));
        thumb(thumb>0.95) = 0.95;
        thumb(thumb<0.075) = 0.075;
        thumb = (1/0.95)*(thumb - min(thumb(:))); % / (max(thumb(:)) - min(thumb(:)));
        if (vis)
          figure(1); clf; imagesc(thumb); axis image; 
          caxis([0 1]);
          colormap(fademapb(256)); drawnow;
        end;
        if (tt==MRNA)
          tname = sprintf('%s/thumbs/%s__%d_mRNA.jpg',exportdir,gene_names{g},cht);
        elseif (tt==PROTEIN)
          tname = sprintf('%s/thumbs/%s__%d_Protein.jpg',exportdir,gene_names{g},cht);
        end;
        imwrite(ceil(255*thumb)+1,fademapb(256),tname);
      end
    end;
  end;
end

fid = fopen([exportdir '/sourcelist.txt'],'w');
for i = 1:length(srclistfinal)
  fprintf(fid,'%s.pce,v5_%s.pce\n',srclistfinal{i},srclistfinal{i});
end

fid = fopen([exportdir '/sourcelist_bygene.txt'],'w');
for gg = 1:numgenes %loop over genes 
  for cht = 4:9
  tt = 1;
      list = srclist_bygene{gg,cht,tt};
      fprintf(fid,'%s - %s - %s:\n',gene_names{gg},channeltype_names{tt},cohort_names{cht})
      for s = 1:length(list)
        fprintf(fid,'   %s\n',list{s});
      end
      fprintf(fid,'\n');
  end
end


if (0)
   cd export/thumbs
   zip ../thumbs.zip *.jpg
   cd ../../pointclouds.cor/
   zip ../export/corfiles.zip *.cor

end
