%
% warp embryo with nuclear coordinates X onto the model with coordinates VV based
% on matching between edges and template given by (match/ematch).  lambda specifies
% the regularization parameter used by the 3D thin-plate spline.  hmap is an 
% interpolated hieghtmap for the morphological template which allows us to lift 
% precise stripe boundary locations specified in cylindrical coordinates into 3D
% space with "sub-nuclear" precision via interpolation over the surface.
%
% charless fowlkes, (c) 2007
%

function [warped,nout] = dp_warp_smooth(X,VV,match,ematch,tplate,slices,lambda,hmap1,qthresh)

    if (nargin < 9)
      qthresh = 40;  %cell atlas value was 10
    end;

    %extract coordinates of matching points
    %in cylindrical projection
    ex = []; et = [];
    tx = []; tt = [];
    for k = 1:length(match)
      ind = find(~isnan(match{k}));
      ex = [ex;ematch{k}(ind)'];
      et = [et;repmat(slices(k),length(ind),1)];
      tx = [tx;tplate{k}.peaks(ind)'];
      tt = [tt;repmat(slices(k),length(ind),1)];
    end;

    %lift points to 3D in order to apply 3D spline
    x = linspace(0,1,200);
    y = linspace(0,1,200);
    tr = interp2(x,y,hmap1,tx,tt,'linear');

    height = sum(X(:,2:3).^2,2).^0.5;
    hmap2 = render_signal(X,height,200,'cubic');
    er = interp2(x,y,hmap2,ex,et,'linear');

    % matching points in 3D using interpolated (sub-nuclear)
    %  coordinates along the stripe boundaries
    P1 = [ex er.*sin(2*pi*et) er.*cos(2*pi*et)];
    P2 = [tx tr.*sin(2*pi*tt) tr.*cos(2*pi*tt)];

    %force the poles to map to each other
    [val,lind1] = min(VV(:,1));
    [val,rind1] = max(VV(:,1));
    [val,lind2] = min(X(:,1));
    [val,rind2] = max(X(:,1));
    P1 = [VV(lind1,:); VV(rind1,:); P1];
    P2 = [X(lind2,:); X(rind2,:); P2];

    %quadratic pruning of matching
    N = size(P1,1); 
    vec = (P1-P2);
    D = dist2(P1,P1);
    Q = dist2(vec,vec) ./ (D + (D==0)); 
    ind = find(sum(Q)<qthresh);
    nout = sum(sum(Q)>=qthresh);
    fprintf('pruning %d outliers of %d leaving %d\n',nout,N,length(ind))

    %TODO: make outlier pruning based on a percentage of the total number of points
    if (nout<200) & (length(ind)>30)
      fprintf('warping with %f inliers\n',length(ind)/N);
      warped = warp3Dtps2(P1(ind,:),P2(ind,:),X,lambda); %proper tps, m=3;
    else
      fprintf('Warning: Too many outliers detected\n');
      warped = [];
    end;

