%
% load in the pointcloud file format into a
%  MATLAB data structure.
%
% charless fowlkes, (c) 2005-2007
%

function pcstruct = load_pc(filename)

pcstruct = [];
if (~fileexists(filename))
  error(sprintf('error opening pc file %s',filename));
end;
fid = fopen(filename,'r');

%parse the header
L = fgetl(fid);
L = eatcomments(fid,L);
while (ischar(L) & findstr(L,'#'))
  L = eatprefix(L);
  [name,value] = parsenvpair(L); 
  pcstruct = setfield(pcstruct,name,value);
  L = fgetl(fid);
  L = eatcomments(fid,L);
end;
fclose(fid);

if ~isfield(pcstruct,'column')
  error('no column field in pc file, parsing aborted');
end;

%now read in the cell data
D = dataread('file',filename,'','commentstyle','shell','delimiter',',');
ncolumns = length(pcstruct.column);

%special colmumns: nuclear geometry
idind = strmatch('id',pcstruct.column);
pcstruct.id = D(:,idind);

remainder = setdiff(1:ncolumns,idind);
if isfield(pcstruct,'cohort_times')
  for i = 1:length(pcstruct.cohort_times)
    pf = sprintf('__%d',i);
    coordind = [strmatch(['x' pf],pcstruct.column,'exact') strmatch(['y' pf],pcstruct.column,'exact') strmatch(['z' pf],pcstruct.column,'exact')];
    pcstruct.X{i} = D(:,coordind);
    normalind = [strmatch(['Nx' pf],pcstruct.column,'exact') strmatch(['Ny' pf],pcstruct.column,'exact') strmatch(['Nz' pf],pcstruct.column,'exact')];
    pcstruct.N{i} = D(:,normalind);
    remainder = setdiff(remainder,[coordind normalind]);
  end
else
  coordind = [strmatch('x',pcstruct.column,'exact') strmatch('y',pcstruct.column,'exact') strmatch('z',pcstruct.column,'exact')];
  pcstruct.X = D(:,coordind);
  normalind = [strmatch('Nx',pcstruct.column,'exact') strmatch('Ny',pcstruct.column,'exact') strmatch('Nz',pcstruct.column,'exact')];
  pcstruct.N = D(:,normalind);
  remainder = setdiff(remainder,[coordind normalind]);
end;

%special columns: neghborhood data
pcstruct.nnbrs = D(:,ncolumns+1);
pcstruct.nbrs = D(:,ncolumns+2:end);

%mild consistency check of neighborhood data
if (sum((sum(pcstruct.nbrs>0,2)-pcstruct.nnbrs)>0)~=0)
  error('neighborhood data is inconsistent with number of neighbors for some row');
  find((sum(pcstruct.nbrs>0,2)-pcstruct.nnbrs)>0)
end;

%remaining per nucleus data
pcstruct.column = pcstruct.column(remainder);
if (isfield(pcstruct,'column_factors'))
  pcstruct.column_factors = pcstruct.column_factors(remainder);
end;
pcstruct.data = D(:,remainder);

%get egglength coordinates
if ( isfield(pcstruct,'translate') & isfield(pcstruct,'rotate1') & isfield(pcstruct,'scale') )
  XX = [pcstruct.X ones(size(pcstruct.X,1),1)]*pcstruct.translate'*pcstruct.rotate1'*pcstruct.scale';
  XX = XX(:,1:3);
  translation = min(XX(:,1)); 
  pcstruct.eggtranslate = [1 0 0 -translation(1); 0 1 0 0; 0 0 1 0; 0 0 0 1];
end;

%combine column_info and column_info_bid
if (isfield(pcstruct,'column_info_bid'))
  len = size(pcstruct.column_info,1);
  for k = len+1:len+size(pcstruct.column_info_bid,1)
    for l = 1:size(pcstruct.column_info,2)
      pcstruct.column_info{k,l} = pcstruct.column_info_bid{k-len,l};
    end;
  end;
  pcstruct = rmfield(pcstruct,'column_info_bid');
end;

%
% deal with transgenes
%
if isfield(pcstruct,'phenotype_string')
  if strcmp(pcstruct.phenotype_string(1:8),'D. melan')
    pcstruct.genotype_id = 1;
    [drop,tgstring] = strtok(pcstruct.phenotype_string,',');
    tgstring = strtok(tgstring,',');
    tgstring = tgstring(2:end);
    for k = 1:size(pcstruct.column_info,1)
      if strcmp(pcstruct.column_info{k,3},'LacZ')
        pcstruct.column_info{k,3} = ['LacZ:' tgstring];
        pcstruct.column_info{k,4} = ['LacZ:' tgstring];
        pcstruct.column_info{k,6} = 'nuclear';
      end
    end
  end
end

%
% skip over lines starting with '##'
%

function L = eatcomments(fid,L)

while (ischar(L) & findstr(L,'##'))
  L = fgetl(fid);
end;

%
% eat leading # and whitespace
%

function L = eatprefix(L)
while(isspace(L(1)) | strcmp(L(1),'#'))
  L = L(2:end);
end;





