%
% normalize signals.  
%
%   find affine transform of each row of the input signal array
%   which minimizes the variance.  based on description from
%   gregor et. al. unpublished communication, 2006
%
% adjusts the overall gain and offset so that the resulting
% mean has the same range as the original
%
% charless fowlkes, (c) 2006-2007
%


% err : sum(sum((ss2-repmat(mean(ss2),N,1)).^2))

function [Snorm,gain,offset] = normalize_signals(S)

  N = size(S,1);

  if (N>1)
    %get the range of the original signals
    original = mean(S);
    range = max(original)-min(original);

    %make the signals zero mean
    o1 = -mean(S')';
    Snorm = S + repmat(o1,1,size(S,2));

    %
    % prune out bad signals
    %

    %intial guess of the gains.
    for k = 1:N
      g0(k) = 1./norm(Snorm(k,:));
    end
    % initial guess for mean
    mu = mean(diag(g0)*Snorm);
    mu = mu ./ norm(mu);
    % given the mean, estimate optimal gains
    g0 = (Snorm*mu')./diag(Snorm*Snorm');

    % ignore those examples with negative gains.
    good = find(g0>0);

    %
    % find optimal mean and gains simultaneously
    %
    % now compute the optimal mean signal for this subset.  
    % SVD gets better values than the alternating 
    % minimization so use it on the good embryos
    Snormgood = Snorm(good,:);
    [u,s,v] = svd(Snormgood*Snormgood');
    uu = Snormgood'*u;
    uu = uu(:,1) ./ sqrt(sum(uu(:,1).^2));

    % svd can swap the sign arbitrarily, make sure
    % uu looks pretty much like mu
    if ((mu*uu)<0)
      uu = -1*uu;
    end

    %find scaling which maximizes projection onto mean signal
    g1 = (Snorm*uu)./diag(Snorm*Snorm');
    bad = find(g1<0);
    g1(bad) = 0;

    % error checking
    ss1 = diag(g1)*Snorm;
    e1 = sum(sum((ss1-repmat(mean(ss1),N,1)).^2));
    fprintf('nz = %d, svd err =%2.2f\n',length(bad),e1);

    %finally scale Snorm
    Snorm = diag(g1)*Snorm;

    % now adjust offset and scale so that 
    % the mean of normalized signals has the
    % same range as the original set of signals
    Save = mean(Snorm);
    o2 = -min(Save);
    g2 = max(Save+o2);
    Snorm = range * ((Snorm+o2) / g2);

    %  multiply out the transform into a single offset and gain.
    %  e.g.
    %
    %    Snorm = range * (diag(g1)*(S + o1)) + o2) / g2;
    %    Snorm = range*(1/g2)*diag(g1)*S + range*((1/g2)*diag(g1)*o1+(o2/g2)));
    %
    % if the calcuation is correct then:
    %
    %   diag(gain)*S + repmat(offset,1,size(S,2)) ==  Snorm
    %
    gain = range*(1/g2).*g1;
    offset = range*((1/g2)*diag(g1)*o1 + (o2/g2));
  elseif (size(S,1)==1)
    Snorm = S;
    gain = 1;
    offset = 0;
  elseif (size(S,1)==0)
    error('empty matrix');
  end;
