% register a given pointcloud onto the atlas.  output the resulting
% set of correspondences to a file.
%
% embryofile - .pce file for individual embryo
% corrfile - file to write correposndence out to
% gene_names - gene_names for the present atlas
% vis - visualize matching process?
%
%
% charless fowlkes, (c) 2007
%

function register_pc(embryofile,corrfile,gene_names,versionstring,vis,forcecoarse);
  addpath edge; %edge extraction module 

  lambda = 2*(10^-3); % warping regularization...selected based on error analysis
  if (nargin <= 5)
    forcecoarse = false;
  end
  if (nargin <= 4)
    vis = false;  % don't visualize matching
  end

  % load in the embryo
  fprintf('loading pointcloud,');
  pc = load_pc(embryofile);

  %
  % static data specifying the stripe location template
  %
  define_constants;
  fprintf('template,');
  try
    load(sprintf('template/%d_stage5shape.mat',MEL));
  catch
    fprintf('\nERROR: no shape model for embryo %s, likely unknown genotype\n',pc.name);
    return;
  end

  try
    load(sprintf('template/%d_dp_templates.mat',MELBCD));
    ftz_mask = imdilate(ftz_mask,ones(5,5));
    missingtemplate = false;
  catch
    fprintf('\nWARNING: no marker gene template for embryo %s\n',pc.name);
    missingtemplate = true;
  end

  % normalize coordinates
  [tX,tN] = transform(pc);
  if ~isfield(pc,'DVrotation')
    %if not rotated, apply a random rotation
    fprintf('applying random rotation');
    theta = 2*pi*rand(1);
    rotate = [1,0,0; 0,cos(theta),sin(theta); 0, -sin(theta), cos(theta)];
    tX = tX*rotate;
  end
  sytox = strmatch(pc.nuclear_stain,pc.column);
  [col,info,zerocount] = getgenechannels(pc,gene_names);
  
  if (size(col,1)==0)
    fprintf('!!!no gene channels, skipping registration\n');
    return
  end

  % extract attenuation offset values 
  offset = zeros(size(info,2),1);
  if ~isempty(pc.attenuation_correct)
    for j = 1:size(info,2)
      ind = strmatch(pc.column_info(info(j),1),pc.attenuation_correct);
      if isempty(ind)
        offset(j) = 0;
      else
        offset(j) = pc.attenuation_offset(ind);
      end;
    end;
  end;

  for j = 1:size(col,2)
    %attenuation correction, set to 1 for the "top" of the embryo
    correction = (pc.data(:,sytox) + offset(j));
    topmost = find((pc.X(:,3)-min(pc.X(:,3)))<2);
    correction = correction / mean(correction(topmost));
    fluor = pc.data(sub2ind(size(pc.data),1:size(pc.data,1),col(:,j)'))';
    expression(:,j) =  fluor ./ correction;
    genes(j) = dict(pc.column_info{info(j),3},gene_names);
    rna(j) = strcmp(pc.column_info{info(j),2},'mRNA');
  end;

  cht = getcohorts(pc);


  if ((cht>3) & (cht<10))
    %
    % compute "nearest" source nucleus for each target nucleus in the atlas
    %
    nearest = [];

    %
    % target average geometry for this cohort/genotype
    %
    VV = V{cht};
    VV(:,1) = VV(:,1)-min(VV(:,1)); VV = VV/max(VV(:,1));

    if (~missingtemplate & ~forcecoarse)
      %
      % pull out the marker gene channel..default to ftz 
      % if both marker genes are present.
      %
      channel = [];
      if ~isempty(dict('ftz',gene_names)) & (any(genes==dict('ftz',gene_names)))
          channel = find((genes==dict('ftz',gene_names)) & (rna==true));
          tplate = template.ftz{cht};
          mask = ftz_mask;
      end;
      
      if (isempty(channel))
        %
        % no marker gene, just do coarse alignment
        %
        fprintf('\nWARNING: embryo %s is missing mRNA marker gene...',pc.name);
        failed = true;
      else
        %
        % marker gene present, lets use it
        %
        marker_expression = expression(:,channel);

        %
        % find corresponding points along ftz/eve stripes
        %
        failed = false;
        fprintf('matching...');
        [match,ematch,qcost] = dp_register_bcdrnai(tX,marker_expression,tplate,slices,mask,0,vis);

        if isempty(match)
          fprintf('matching failed...');
          failed = true;
        else
          height = sum(VV(:,2:3).^2,2).^0.5;
          hmap = render_signal(VV,height,200,'cubic');
          fprintf('warping...');
          warped = dp_warp_smooth(tX,VV,match,ematch,tplate,slices/200,lambda,hmap);
          if isempty(warped)
            fprintf('warping failed...');
            failed = true;
          else
            fprintf('corresponding\n');
            [dist,nearest] = mindist2(warped,VV);
          end
        end
      end
    else
      %
      % no template, just do coarse alignment
      %
      if (~forcecoarse)
        fprintf('\nWARNING: missing template...');
      end
      failed = true;
    end 

    %
    % failure mode if either missing marker gene, template or warping failed
    %  in this case just do nearest neighbor assignment based on coarse align.
    if (failed|forcecoarse)
      fprintf('defaulting to coarse alignment\n');
      [dist,nearest] = mindist2(tX,VV);
      qcost = inf;
    end;

  elseif (cht==3)
    %
    % special case to handle 4d...just use 5:0-3% geometry
    %
    fprintf('!!stage 4d!!, mapping to 5:0 atlas, ');
    fprintf('defaulting to coarse alignment\n');
    failed = true;

    VV = V{4};
    VV(:,1) = VV(:,1)-min(VV(:,1));
    VV = VV/max(VV(:,1));

    [dist,nearest] = mindist2(tX,VV);
    qcost = inf;
  else
    fprintf('\nERROR: embryo %s is outside atlas stage range...skipping\n',pc.name);
    return;
  end;

  %
  % nearest should exist at this point, either via coarse or fine alignment
  %
  if (isempty(nearest))
    error('assertion failure, nearest is empty')
  end;

  matchmat = sparse(1:length(nearest),nearest,ones(size(nearest)),size(VV,1),size(tX,1));
  % TODO: for now we just have 1 match per row so we don't need to normalize.
  %   this normalization would be really slow since it builds a dense matrix
  %matchmat = matchmat ./ repmat(sum(matchmat,2),1,size(matchmat,2));
  fprintf('saving result\n');
  save_corr(corrfile,matchmat,sprintf('atlas_%s',versionstring),pc.name,sum(dist).^0.5,qcost,failed);

